import 'dart:async';
import 'dart:convert';

import 'package:filesize/filesize.dart';
import 'package:logger/logger.dart';
import 'package:roundtripper/roundtripbuilders/request_body_convert.dart';
import 'package:roundtripper/roundtripbuilders/request_log.dart';
import 'package:roundtripper/roundtripbuilders/throws_not_2xx_error.dart';
import 'package:roundtripper/roundtripper.dart';
import 'package:test/test.dart';

var logger = Logger(
  printer: LogfmtPrinter(),
);

var httpBinEndpoint = "https://httpbin.org";

void main() async {
  var c = Client(roundTripBuilders: [
    ThrowsNot2xxError(),
    RequestBodyConvert(),
    RequestLog(logger),
  ]);

  test("get", () async {
    var resp = await c.fetch(Request.uri(
      "$httpBinEndpoint/anything",
      queryParameters: {
        "int": 1,
        "slice": [1, 2],
      },
      headers: {
        "x-int": 1,
        "x-slice": [1, 2],
      },
    ));

    expect(resp.statusCode, 200);

    var json = await resp.json();

    expect(json["url"], "$httpBinEndpoint/anything?int=1&slice=1&slice=2");
    expect(json["headers"]["X-Int"], "1");
    expect(json["headers"]["X-Slice"], "1, 2"); // ?
  });

  test("post", () async {
    var resp = await c.fetch(Request.uri(
      "$httpBinEndpoint/anything",
      method: "POST",
      body: {
        "a": 1,
        "b": "s",
      },
    ));

    expect(resp.statusCode, 200);

    var json = await resp.json();

    expect(jsonDecode(json["data"]), {
      "a": 1,
      "b": "s",
    });

    expect(json["headers"]["Content-Type"], "application/json; charset=utf-8");
  });

  test("download & progress", () async {
    var resp = await c.fetch(Request.uri(
      "$httpBinEndpoint/bytes/102400",
      method: "GET",
    ));

    expect(resp.statusCode, 200);

    var complete = 0;
    var len = resp.contentLength;

    var f = StreamController(sync: true);

    f.stream.listen((event) {});

    await f.addStream(
      resp.responseBody.transform(
        StreamTransformer<List<int>, List<int>>.fromHandlers(
          handleData: (data, sink) {
            sink.add(data);
            complete += data.length;
            logger.i("${filesize(complete)} / ${filesize(len)}");
          },
        ),
      ),
    );

    f.close();
  });

  test("throws not 2xx error", () async {
    await expectLater(
      c.fetch(Request.uri("$httpBinEndpoint/status/401")),
      throwsA(
        (e) => e is ErrorResponse && e.statusCode == 401,
      ),
    );
  });

  test("cancel", () async {
    final token = CancelToken();

    Timer(const Duration(milliseconds: 1), () {
      token.cancel("cancel");
    });

    await expectLater(
      c.fetch(Request.uri("$httpBinEndpoint/status/200", cancelToken: token)),
      throwsA(
        (e) =>
            e is ErrorResponse &&
            CancelToken.isCancel(e) &&
            e.reason == "cancel",
      ),
    );
  });
}
